"""A little testing config."""

import hashlib
import os

from spin.config import Config, Device, Page, Param

config = Config()
config.server = '0.0.0.0:8000'
config.control_url = 'http://{host}:8080/'
config.plugins = ['plugin.py']
config.nav_bar = True

tg = 'tango://localhost:10000/test/sim'
sec = 'secop://localhost:10767'
nic = 'nicos://localhost/nicos'
plc = f'pils+sim://{os.path.dirname(__file__)}/plc.py'  # noqa: PTH120
config.devices = {
    'field': Device(device=f'{tg}/motor',
                    params=[Param('ramp', 5.0), 'target']),
    'coil_a': Device(device=f'{tg}/sensor'),
    'coil_b': Device(device=f'{tg}/sensor'),
    'coldhead_stage1': Device(device=f'{tg}/sensor'),
    'coldhead_stage2': Device(device=f'{tg}/sensor'),
    'fsm': Device(device=f'{tg}/digitaloutput'),
    'valve': Device(device=f'{tg}/discreteoutput',
                    confirm='Do not open the valve without proper training!'),
    'valvetext': Device(device=f'{tg}/discreteoutput', options={'enum': True}),
    'cryo_label': Device(device=f'{sec}/label'),
    'cryo_tc': Device(device=f'{sec}/tc1'),
    'cryo_ts': Device(device=f'{sec}/ts'),
    'cryo_switch': Device(device=f'{sec}/heatswitch', options={'enum': True}),
    'net_rx': Device(device=f'{nic}/net_rx'),
    'net_tx': Device(device=f'{nic}/net_tx'),
    'reactor': Device(device=f'{nic}/reactorpower'),
    'plc_mot': Device(device=f'{plc}#po32', params=['Speed', 'target'],
                      options={'table': 'table2'}),
    'plc_inp': Device(device=f'{plc}#di32', options={'enum': True}),
    'plugin': Device(device='demoplugin://some/device'),
}
config.pages = {
    'default': Page('default.html', title='Spin demo page'),
    'subpage': Page('default.html', title='Subpage', parent='default'),
    'locked':  Page('default.html', title='Locked subpage', parent='default',
                    auth=[('admin', hashlib.sha256(b'admin').hexdigest())]),
    'view':    Page('default.html', readonly=True, title='Spin demo (r/o)',
                    quick_icons=False),
    'table':   Page('table.html', title='PILS table'),
    'secret':  Page('secret.html', title='Very secret', hidden=True,
                    auth=[('admin', hashlib.sha256(b'admin').hexdigest()),
                          ('heiner', hashlib.sha256(b'heiner').hexdigest())]),
}



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Config handling logic."""

import asyncio
from dataclasses import dataclass, field


@dataclass
class Param:
    """Configuration for one parameter to be polled."""

    name: str
    # None means to use the device's default
    pollinterval: float = None
    # set internally
    poll_every: int = 1


@dataclass
class Device:
    """Configuration for one device to be polled."""

    device: str
    pollinterval: float = 1.0
    params: list = None
    confirm: str = None
    options: dict = field(default_factory=dict)


@dataclass
class Page:
    """Configuration for a page served by the frontend."""

    file: str
    title: str = None
    auth: list[tuple] = None
    hidden: bool = False
    readonly: bool = False
    nav_bar: bool = None
    quick_icons: bool = True
    parent: str = None


@dataclass
class Config:
    """The main configuration."""

    server: str = 'localhost:8000'
    plugins: list[str] = field(default_factory=list)
    control_url: str = None
    devices: dict[str, Device] = None
    pages: dict[str, Page] = None
    dark_theme: bool = False
    nav_bar: bool = False
    quick_icons: bool = True


def get_config(args):
    conffile = args.confdir / 'conf.py'
    namespace = {'__file__': str(conffile)}
    code = conffile.read_bytes()
    exec(code, namespace)  # pylint: disable=exec-used  # noqa: S102
    return namespace['config']


async def config_watcher(args, callback):
    """Watch the configuration file for changes."""
    conffile = args.confdir / 'conf.py'
    last_mtime = conffile.stat().st_mtime
    while True:
        await asyncio.sleep(1)
        try:
            mtime = conffile.stat().st_mtime
        except OSError:
            pass
        else:
            if mtime != last_mtime:
                return await callback()



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Main spin package."""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__)
except importlib.metadata.PackageNotFoundError:
    try:
        from setuptools_scm import get_version  # pylint: disable=import-error
        __version__ = get_version(root='..', relative_to=__file__)
        del get_version
    except Exception:  # noqa: BLE001
        __version__ = 'unknown'



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""SECoP backend."""

import json
from collections import defaultdict

from .backend import State
from .device import AsyncUpdateDevice
from .utils import StreamConnection


def to_state(status):
    status, statustext = status
    if status < 0:
        return State.UNKNOWN, ''
    if status < 100:
        return State.DISABLED, statustext
    if status < 200:
        return State.OK, statustext
    if status < 300:
        return State.WARN, statustext
    if status < 400:
        return State.BUSY, statustext
    if status < 500:
        return State.ERROR, statustext
    return State.UNKNOWN, ''


# We want to open only one connection per node, so save them here
NODES = {}


class SECNode(StreamConnection):
    """Device representing a SEC node connection."""

    async def _synchronize(self):
        self.log.info('connected to node')
        await self._send(b'*IDN?')
        proto = await self._receive()
        if ',SECoP,' not in proto:
            raise RuntimeError('not a SECoP node, IDN = {proto!r}')
        await self._send(b'describe')
        description_msg = await self._receive()
        _, _, description = self._decode(description_msg)
        # check if identifiers are in description and set all units
        await self._set_units(description)
        self.log.info('waiting for initial updates')
        await self._send(b'activate')
        unseen = set(self.devices)
        while True:
            msg = self._decode(await self._receive())
            if msg[0] == 'active':
                break
            await self._handle_message(msg)
            unseen.discard(msg[1])
        if unseen:
            self.log.warning('not all initial updates from SEC node '
                             'received (missing %s)', unseen)

    async def _set_units(self, descr):
        for name, devs in self.devices.items():
            for dev in devs:
                mod, _, param = name.partition(':')
                if param == 'status':
                    continue
                try:
                    modules = descr['modules']
                except KeyError as e:
                    raise RuntimeError('Malformed description message!') from e

                try:
                    datainfo = modules[mod]['accessibles'][param]['datainfo']
                    dev.set_datainfo(param, datainfo)
                except KeyError:
                    # pylint: disable=raise-missing-from
                    dev.log.error('accessible %s not present on SEC node', name)
                    await dev.send(State.ERROR,
                                   'Module/param not present on SEC node')

    def _decode(self, line):
        action, ident, data = [*line.strip().split(' ', 2), '', ''][:3]
        data = {} if data == '' else json.loads(data)
        return action, ident, data

    async def _handle_message(self, msg):
        action, ident, data = msg
        if ident not in self.devices:
            return
        for dev in self.devices[ident]:
            await dev.update(action, ident, data)

    def add_device(self, device):
        mod = device.device
        self.devices[f'{mod}:value'].append(device)
        # we always want the status
        self.devices[f'{mod}:status'].append(device)
        for par in device.extra_params or []:
            self.devices[f'{mod}:{par}'].append(device)

    async def run_cmd(self, dev, cmd):
        await self._send(f'do {dev}:{cmd}'.encode())

    async def write(self, dev, param, value):
        await self._send(f'change {dev}:{param} {value}'.encode())


class Device(AsyncUpdateDevice):
    """A device representing a polled secop module.

    The polling happens in the SEC node, this is kept for callbacks on actions.
    Device url should look like: 'secop://<host>:<port>/<module>'
    """

    _connclass = SECNode

    def __init__(self, name, devconf, uri, backend, log):
        super().__init__(name, devconf, uri, backend, log)
        self._input_value_type = {}
        self._useenum = devconf.options.get('enum', False)
        self._enums = {}
        self._units = defaultdict(str)

    def set_datainfo(self, param, info):
        self.log.info('datainfo is set')
        self._units[param] = info.get('unit', '')
        if param == 'value':
            self._unit = self._units[param]
        if info.get('type') == 'double':
            mini = info.get('min', -1e308)
            maxi = info.get('max', 1e308)
            self._input_value_type[param] = ['float', mini, maxi]
        elif info.get('type') == 'int':
            mini = info.get('min', -2147483648)
            maxi = info.get('max', 2147483647)
            self._input_value_type[param] = ['int', mini, maxi]
        elif info.get('type') == 'bool':
            self._input_value_type[param] = ['int', 0, 1]
        elif info.get('type') == 'enum':
            members = info.get('members', {})
            if self._useenum:
                rev_members = {v: k for k, v in members.items()}
                self._enums[param] = members, rev_members
                self._input_value_type[param] = ['choice', list(members)]
            else:
                self._input_value_type[param] = ['int', 0, len(members) - 1]
        elif info.get('type') == 'string':
            self._input_value_type[param] = ['str']

    async def update(self, action, ident, data):
        _, _, param = ident.partition(':')
        value = data[0]
        if action == 'update':
            if param == 'status':
                self.cache['status'] = to_state(value)
                state, state_str = self.cache['status']
                await self.send(state, state_str, self.cache['value'])
            elif param == 'value':
                if 'value' in self._enums and value in self._enums['value'][1]:
                    value = self._enums[param][1][value]
                self.cache['value'] = value
                state, state_str = self.cache['status']
                await self.send(state, state_str, value)
            else:
                if param in self._enums and value in self._enums[param][1]:
                    value = self._enums[param][1][value]
                self.cache[param] = value
                state, state_str = self.cache['status']
                await self.send_param(param, value, self._units[param])

    async def input_value_type(self, param):
        return self._input_value_type.get(param, ['float', -1e308, 1e308])

    async def action_stop(self):
        await self._conn.run_cmd(self.device, 'stop')

    async def action_reset(self):
        # TODO: check if reset exists
        await self._conn.run_cmd(self.device, 'reset')

    async def action_set(self, param, value):
        if param == 'value':
            param = 'target'
        if (members := self._enums.get(param)) and \
           (value in members[0]):
            value = members[0][value]
        await self._conn.write(self.device, param, json.dumps(value))

    async def action_run(self, command):
        await self._conn.run_cmd(self.device, command)

    async def action_upload(self, filename, file, target=None):
        pass



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Plugin for handling Lakeshore calibration curves."""

from typing import NamedTuple

from spin.backend import Reply, Request
from spin.device import BaseDevice
from spin.utils import decode_base64_file

try:
    import tango
    import tango.asyncio
except ImportError:
    tango = None

# Lakeshore curve lists (set 0 to not use one on the sensor)
# Lakeshore,    fixed,  user,   unused
# LS 211,       1-20,   21-21,  -
# LS 218,       1-9,    21-28,  10-20
# LS 224,       1-20,   21-59,  -
# LS 332,       1-20,   21-41,  -
# LS 335,       1-20,   21-59,  -
# LS 336,       1-20,   21-59,  -
# LS 340,       1-20,   21-60,  -
# LS 350,       1-20,   21-59,  -
# LS 372,       1-20,   21-59,  -

# Model -> (Max idx predef curve, Max idx user curve)
CURVES = {
    'MODEL211': (20, 21),
    'MODEL218': (9, 28),
    'MODEL218S': (9, 28),
    'MODEL224': (20, 59),
    'MODEL332': (20, 41),
    'MODEL335': (20, 59),
    'MODEL336': (20, 59),
    'MODEL340': (20, 60),
    'MODEL350': (20, 59),
    'MODEL372': (20, 59),
}

FORMAT = {
    1: 'mV/K',
    2: 'V/K',
    3: 'Ohm/K (linear)',
    4: 'log(Ohm)/K',
    5: 'log(Ohm)/log(K)',
    7: 'Ohm/K (cubic spline)',
}

COEFFICIENT = {
    1: 'negative',
    2: 'positive',
}

HEADER_FIELDS = {
    'sensor model': 'name',
    'serial number': 'serial',
    'data format': 'fmt',
    'setpoint limit': 'limit',
    'temperature coefficient': 'coeff',
    # number of breakpoints
}

# TODO: only wait 0.05s or longer to be safe for the polling?
WAIT_TIME = 0.05


def to_340_file(curve, points):
    # TODO: Sensor Model is name?, spacing
    out = f'Sensor Model: {curve.name}\n'
    out += f'Serial Number: {curve.serial}\n'
    out += f'Data Format: {curve.fmt}\n'
    out += f'SetPoint Limit: {curve.limit}\n'
    out += f'Temperature coefficient: {curve.coeff}\n'
    out += f'Number Of Breakpoints: {len(points)}\n'
    out += '\n No.  Units    Temperature (K)\n\n'
    for i, point in enumerate(points):
        unit, temp = point.split(',')
        out += f'{i+1:>3} {unit:>6}    {temp:>6}\n'
    return out


def from_340_file(idx, file):
    points = []
    fields = {'idx': idx}
    for line in file.splitlines():
        if not line:
            continue
        if ':' in line:
            key, value = line.split(':')
            key = key.strip().lower()
            if key in HEADER_FIELDS:
                fields[HEADER_FIELDS[key]] = value.strip()
        else:
            if line.strip().lower().startswith('no.'):
                continue
            i, x, y = line.split()
            points.append((int(i), x, y))
    try:
        curve = Curve(**fields)
    except TypeError as e:
        raise ValueError('.340 file header does not contain all expected '
                         'fields') from e
    return curve, points


class Curve(NamedTuple):
    """Represents a Lakeshore curve header."""

    idx: int
    name: str
    serial: str
    fmt: int
    limit: float
    coeff: float

    def to_dict(self):
        res = self._asdict()
        coeff = res['coeff']
        res['coeff'] = COEFFICIENT.get(coeff, coeff)
        fmt = res['fmt']
        res['fmt'] = FORMAT.get(fmt, fmt)
        return res

    def to_str(self):
        return (f'{self.idx},{self.name},{self.serial},'
                f'{self.fmt},{self.limit},{self.coeff}')


def parse_curve(idx, curve):
    curve = curve.removeprefix('CRVHDR ')
    name, serial, fmt, limit, coeff = curve.split(',')  # TODO: comma in name?
    return Curve(idx, name.strip(), serial.strip(),
                 int(fmt), float(limit), int(coeff))


class LakeshoreCurveDeviceTango(BaseDevice):
    # pylint: disable=abstract-method
    """Lakeshore Curve Up- and Download."""

    # TODO: DOCUMENT!

    _comm = None

    def init(self):
        if not tango:
            raise RuntimeError('PyTango is required to use this plugin')
        self.run_task(self._setup())

    async def _setup(self):
        self.log.info('setting up connection for lakeshore plugin')
        self._comm = await tango.asyncio.DeviceProxy(
                self._devconf.device.replace('lsfile+tango', 'tango'))

    async def _communicate(self, msg):
        return (await self._comm.command_inout('MultiCommunicate',
                                               [[-WAIT_TIME], [msg]]))[0]

    async def _tell(self, msg):
        await self._comm.command_inout('MultiCommunicate',
                                       [[WAIT_TIME], [msg]])

    async def handle_request(self, request: Request):
        if request.action == 'scan':
            self.run_task(self._read_headers(request.connection))
        elif request.action == 'download':
            self.run_task(self._download_curve(request.connection,
                                               request.data))
        elif request.action == 'upload':
            self.run_task(
                self._update_curve(request.connection,
                                   request.data['idx'],
                                   request.data['contents']),
            )
        else:
            self.log.warning('got unknown request %r', request)

    async def _read_headers(self, conn):
        async with self.error_guard(
                conn, msg='Error during curve scanning:', reset_progress=True):
            self.log.info('scanning headers')
            await self.send_progress(conn, 0, 'Scanning curves...')
            await self.send_reply(Reply(conn, self._name, None,
                                        'plugin-lakeshore',
                                        {'reason': 'reset'}))
            model = (await self._communicate('*IDN?')).split(',')[1]
            # TODO: error when unknown device
            max_predef, max_user = CURVES.get(model, (0, 0))
            curves_to_scan = [*range(1, max_predef+1), *range(21, max_user+1)]
            for i, curve_idx in enumerate(curves_to_scan):
                response = await self._communicate(f'CRVHDR? {curve_idx}')
                curve = parse_curve(curve_idx, response)
                data = {
                    'reason': 'curve-found',
                    'curve': curve.to_dict(),
                }
                await self.send_reply(Reply(conn, self._name, None,
                                            'plugin-lakeshore', data))
                await self.send_progress(conn, (i+1) / len(curves_to_scan),
                                         'Scanning curves...')
            self.log.info('scanning headers done')

    async def _download_curve(self, conn, curveidx):
        async with self.error_guard(conn, msg='Error downloading curve:'):
            points = []
            self.log.info('downloading curve %d', curveidx)
            await self.send_progress(conn, 0, 'Reading curves from lakeshore')
            curve = parse_curve(curveidx,
                                await self._communicate(f'CRVHDR? {curveidx}'))
            for i in range(1, 201):
                response = await self._communicate(f'CRVPT? {curveidx},{i}')
                await self.send_progress(conn, i/200,
                                         f'Reading point {i} from lakeshore')
                point = (*map(float, response.split(',')[:2]),)
                if point == (0.0, 0.0):
                    await self.send_progress(conn, 1)
                    break
                points.append(response)
            file = to_340_file(curve, points)
            data = {
                'reason': 'curve-ready',
                'curve': curveidx,
                'curvename': curve.name,
                'file': file,
            }
            await self.send_reply(Reply(conn, self._name, None,
                                        'plugin-lakeshore', data))
            self.log.info('downloading curve done')

    async def _update_curve(self, conn, idx, file):
        async with self.error_guard(
                conn, msg='Error while uploading curve:', reset_progress=True):
            decoded = decode_base64_file(file).decode('utf-8')
            curve, points = from_340_file(idx, decoded)
            self.log.info('updating curve %d', idx)
            total = len(points) + 1
            await self.send_progress(conn, 0, f'Uploading curve: 0/{total}')
            await self._tell(f'CRVHDR {curve.to_str()}')
            for (i, x, y) in points:
                await self._tell(f'CRVPT {idx},{i},{x},{y}')
                await self.send_progress(conn, i/total,
                                         f'Uploading curve: {i}/{total}')
            await self._tell(f'CRVPT {idx},{total},0.0,0.0')
            await self.send_progress(conn, 1)
            self.log.info('updating curve done')


def init_plugin(_conf, _loop):
    return {
        'protos': {
            'lsfile+tango': LakeshoreCurveDeviceTango,
        },
        'assets': ['lakeshore.js', 'lakeshore.css'],
    }



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Readonly NICOS backend."""

import re
from ast import (
    Add,
    BinOp,
    Call,
    Constant,
    Dict,
    List,
    Name,
    Set,
    Sub,
    Tuple,
    UnaryOp,
    USub,
    parse,
)

from .backend import State
from .device import AsyncUpdateDevice
from .utils import StreamConnection

DEFAULT_CACHE_PORT = 14869

OP_TELL = '='
OP_ASK = '?'
OP_WILDCARD = '*'
OP_SUBSCRIBE = ':'
OP_UNSUBSCRIBE = '|'
OP_TELLOLD = '!'
OP_LOCK = '$'
OP_REWRITE = '~'
END_MARKER = '###'

OPKEYS = OP_TELL + OP_ASK + OP_WILDCARD + OP_SUBSCRIBE + OP_UNSUBSCRIBE + \
    OP_TELLOLD + OP_LOCK + OP_REWRITE


# regular expression matching a cache protocol message
# pylint: disable=consider-using-f-string
msg_pattern = re.compile(fr'''
    ^ (?:
      \s* (?P<time>\d+\.?\d*)?                   # timestamp
      \s* (?P<ttlop>[+-]?)                       # ttl operator
      \s* (?P<ttl>\d+\.?\d*(?:[eE][+-]?\d+)?)?   # ttl
      \s* (?P<tsop>@)                            # timestamp mark
    )?
    \s* (?P<key>[^{OPKEYS}]*?)                     # key
    \s* (?P<op>[{OPKEYS}])                        # operator
    \s* (?P<value>[^\r\n]*?)                     # value
    \s* $
    ''', re.VERBOSE)


STATE_MAP = {
    200: State.OK,
    210: State.WARN,
    220: State.BUSY,
    230: State.ERROR,  # NOTREACHED
    235: State.DISABLED,
    240: State.ERROR,
    999: State.UNKNOWN,
}


_safe_names = {'None': None, 'True': True, 'False': False,
               'inf': float('inf'), 'nan': float('nan')}


def ast_eval(node):
    # copied from Python 2.7 ast.py, but added support for float inf/-inf/nan
    def _convert(node):
        if isinstance(node, Constant):
            return node.value
        if isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        if isinstance(node, List):
            return list(map(_convert, node.elts))
        if isinstance(node, Dict):
            return {_convert(k): _convert(v) for k, v
                    in zip(node.keys, node.values)}
        if isinstance(node, Set):
            return frozenset(map(_convert, node.elts))
        if isinstance(node, Name) and node.id in _safe_names:
            return _safe_names[node.id]
        if isinstance(node, UnaryOp) and \
           isinstance(node.op, USub) and \
           isinstance(node.operand, Name) and \
           node.operand.id in _safe_names:
            return -_safe_names[node.operand.id]
        if isinstance(node, UnaryOp) and \
           isinstance(node.op, USub) and \
           isinstance(node.operand, Constant):
            return -node.operand.value
        if isinstance(node, BinOp) and \
           isinstance(node.op, (Add, Sub)) and \
           isinstance(node.right, Constant) and \
           isinstance(node.right.value, complex) and \
           isinstance(node.left, Constant) and \
           isinstance(node.left.value, (int, float)):
            left = node.left.value
            right = node.right.value
            if isinstance(node.op, Add):
                return left + right
            return left - right
        if isinstance(node, Call) and \
           isinstance(node.func, Name) and \
           node.func.id == 'cache_unpickle':
            return None  # we don't support pickled values
        raise ValueError(f'malformed literal string with {node}')
    return _convert(node)


def cache_load(entry):
    try:
        # parsing with 'eval' always gives an ast.Expression node
        expr = parse(entry, mode='eval').body
        return ast_eval(expr)
    except Exception as err:
        raise ValueError(
            f'corrupt cache entry: {entry!r} ({err})') from err


class NicosCache(StreamConnection):
    """Device representing a NICOS cache connection."""

    default_port = DEFAULT_CACHE_PORT

    async def _synchronize(self):
        self.log.info('connected to cache')
        await self._send(f'@{OP_WILDCARD}\n@{OP_SUBSCRIBE}\n'
                         f'{END_MARKER}{OP_ASK}'.encode())
        end = f'{END_MARKER}{OP_TELLOLD}\n'
        unseen = set(self.devices)
        updates = []
        devs = set()
        while True:
            line = await self._receive()
            if line == end:
                break
            msg = self._decode(line)
            if msg[4].endswith('/unit'):
                if (valuekey := msg[4][:-5] + '/value') in self.devices:
                    for dev in self.devices[valuekey]:
                        dev.set_unit(cache_load(msg[6]))
            elif msg[4] in self.devices:  # cache updates until units set
                updates.append(msg)
            unseen.discard(msg[4])
        for key in unseen:
            for dev in self.devices[key]:
                dev.log.error('key %s not present in cache', key)
                await dev.send(State.ERROR,
                               'Device not present in NICOS cache')
        for msg in updates:
            for dev in self.devices[msg[4]]:
                if dev not in devs:
                    devs.add(dev)
                    dev.log.info('device got initial update')
            await self._handle_message(msg)

    def _decode(self, line):
        return msg_pattern.match(line).groups()

    async def _handle_message(self, msg):
        *_, key, op, value = msg
        if key not in self.devices:
            return
        value = cache_load(value) if value else None
        for dev in self.devices[key]:
            await dev.update(key, op, value)

    def add_device(self, device):
        key = device.device
        self.devices[f'{key}/value'].append(device)
        # we always want the status
        self.devices[f'{key}/status'].append(device)
        for par in device.extra_params or []:
            self.devices[f'{key}/{par}'].append(device)


class Device(AsyncUpdateDevice):
    """Abstraction for NICOS devices.

    Device url should look like: 'nicos://<host>:[<port>]/<device>',
    with devicepath being the devices slash-separated name in the nicos cache.
    If host is not given, the default NICOS cache port is used.
    """

    _connclass = NicosCache

    def set_unit(self, unit):
        self._unit = unit

    async def update(self, key, op, value):
        _, _, param = key.rpartition('/')
        if op == OP_TELL:
            if param == 'status':
                nicos_state, state_str = value
                state = STATE_MAP.get(nicos_state, State.UNKNOWN)
                self.cache['status'] = (state, state_str)
                await self.send(state, state_str, self.cache['value'])
            elif param == 'value':
                self.cache['value'] = value
                state, state_str = self.cache['status']
                await self.send(state, state_str, value)
            else:
                self.cache[param] = value
                state, state_str = self.cache['status']
                await self.send_param(param, value, '')
        elif op == OP_TELLOLD and param == 'status':
            self.cache['status'] = (State.UNKNOWN, 'expired')
            await self.send(State.UNKNOWN, 'expired', self.cache['value'])

    async def handle(self, request):
        raise NotImplementedError('NICOS devices cannot be used interactively')



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Backend code for reading values and sending the updates to the frontend."""

import asyncio
import enum
import importlib
import json
import socket
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from urllib.parse import urlparse

import mlzlog

from . import __version__
from .config import Config, Param


class State(str, enum.Enum):  # later: use StrEnum
    """Device states."""

    OK = 'ok'
    WARN = 'warn'
    BUSY = 'busy'
    DISABLED = 'disabled'
    ERROR = 'error'
    UNKNOWN = 'unknown'


@dataclass
class Update:
    """Poll event sent to all connected frontends."""

    key: str
    unit: str
    state: State
    state_str: str
    value: object

    def serialize(self):
        return json.dumps({
            'key': self.key,
            'unit': self.unit,
            'state': self.state.value,
            'state_str': self.state_str,
            'value': self.value,
        })


@dataclass
class Request:
    """Request from a frontend."""

    connection: object = field(repr=False)
    key: str
    transform: object
    action: str
    intent: str
    data: object

    @staticmethod
    def unserialize(connection, data):
        data = json.loads(data)
        return Request(connection,
                       data['key'],
                       data['transform'],
                       data['action'],
                       data['intent'],
                       data['data'])


@dataclass
class Reply:
    """Reply to a Request sent to a specific frontend."""

    connection: object
    key: str
    transform: object
    action: str
    data: object = None

    def serialize(self):
        return json.dumps({
            'key': self.key,
            'transform': self.transform,
            'action': self.action,
            'data': self.data,
        })


# maps Device URI schemes to the module that implements the device
BUILTIN_PROTOS = {
    'tango': 'spin.tango',
    'secop': 'spin.secop',
    'nicos': 'spin.nicos',
    'pils+ads': 'spin.pils',
    'pils+modbus': 'spin.pils',
    'pils+tango': 'spin.pils',
    'pils+sim': 'spin.pils',
}


class Backend:
    """Collection of polling tasks for the devices."""

    def __init__(self, loop, connections, log):
        self.log = log
        self._loop = loop
        self._connections = connections
        self._tasks = set()
        self._devices = {}
        self._last_updates = {}  # last update for each key
        self.protos = BUILTIN_PROTOS.copy()  # this can be extended by plugins

    def start(self, conf: Config, asset_list: list[Path]):
        for name, device in conf.devices.items():
            parsed = urlparse(device.device)
            if parsed.scheme not in self.protos:
                raise ValueError(f'{device.device}: unknown protocol')
            fixup_params(device)
            devcls = self.protos[parsed.scheme]
            if isinstance(devcls, str):
                # import the module if it's a string
                devcls = importlib.import_module(devcls).Device
            log = mlzlog.getLogger(parsed.scheme).getChild(name)
            dev = self._devices[name] = devcls(name, device, parsed, self, log)
            for asset in map(Path, dev.needed_assets()):
                if asset not in asset_list:
                    asset_list.append(asset)
            try:
                dev.init()
            except Exception as err:
                dev.log.exception('while initializing device: %r', err)

        self.run_task(self.periodic())

    def run_task(self, coro):
        task = self._loop.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)

    async def handle_request(self, request):
        devname = request.key.partition('.')[0]
        if not (dev := self._devices.get(devname)):
            self.log.warning('unknown device in request: %r', devname)
            return
        dev.log.debug('handling request: %r', request)
        try:
            reply = await dev.handle_request(request)
        except Exception as err:
            dev.log.exception('error handling request')
            exc_str = dev.format_exception(err)
            await self.send_reply(Reply(
                request.connection,
                request.key,
                request.transform,
                'error',
                f'Could not execute action: {exc_str}',
            ))
        else:
            if reply is not None:
                await self.send_reply(reply)

    async def send_reply(self, reply):
        """Send data to one connected websocket."""
        data = reply.serialize()
        try:
            await reply.connection.send_str(data)
        except ConnectionResetError:
            self.log.warning('error sending reply to %s', reply.connection)

    async def send_all_updates(self, conn):
        """Send all cached data to a new websocket."""
        for data in self._last_updates.values():
            await conn.send_str(data)

    async def send_update(self, update: Update):
        """Send data to all connected websockets."""
        data = update.serialize()
        self._last_updates[update.key] = data
        for conn in self._connections:
            try:
                await conn.send_str(data)
            except ConnectionResetError:
                pass  # socket is being closed down

    async def update_common_info(self):
        fqdn = socket.getfqdn()
        state = State.OK
        if fqdn.startswith('localhost'):
            state = State.WARN
        common = {
            'hostinfo': fqdn,
            'version': f'Connected to Spin {__version__}',
        }
        await self.send_update(Update('__common__', '', state, '', common))

    async def periodic(self):
        while True:
            await self.update_common_info()
            await asyncio.sleep(10)

    async def notify_restart(self):
        """Best-effort notification that the server is restarting."""
        for conn in self._connections:
            with suppress(Exception):
                await conn.send_str(Reply(None, '__common__',
                                          None, 'restart').serialize())


def fixup_params(devconf):
    new_list = []
    for param in devconf.params or []:
        paramcfg = Param(name=param) if not isinstance(param, Param) else param
        poll_every = max(paramcfg.pollinterval // devconf.pollinterval, 1) \
            if paramcfg.pollinterval else 1
        paramcfg.poll_every = poll_every
        new_list.append(paramcfg)
    devconf.params = new_list



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Backend server."""

import argparse
import asyncio
import html
import signal
import socket
import sys
from hashlib import sha256
from pathlib import Path

import mlzlog
from aiohttp import BasicAuth, WSMsgType, hdrs, web

from .backend import Backend, Request
from .config import config_watcher, get_config
from .utils import error_fallback, extract_addr

ASSET_DIR = Path(__file__).parent / 'assets'

# Add provisions for development usage
rootdir = Path(__file__).parent.parent.resolve()
if (rootdir / '.git').exists():  # noqa: SIM108
    default_cfgdir = rootdir / 'demo'
else:  # pragma: no cover
    default_cfgdir = Path('/etc/spin')


def sanitize_path(path):
    """Make sure that the given path is a safe relative path."""
    path = Path(path)
    if path.is_absolute() or '..' in path.parts:
        raise ValueError(f'invalid path {path}')
    return path


class SiteHandler:
    """Handlers for Http and Websocket connections."""

    def __init__(self, conf, conf_root, log):
        self.log = log
        self.connections = set()
        self._asset_root = ASSET_DIR
        self._conf_root = conf_root
        self._conf = conf
        self._backend = None
        self.builtin_assets = [  # can be extended by backends
            Path('pico.min.css'),
            Path('keyboard.css'),
            Path('spin.css'),
            Path('sprintf.js'),
            Path('keyboard.js'),
            Path('filesaver.js'),
            Path('luxon.js'),
            Path('hammer.js'),
            Path('chart.js'),
            Path('chart-luxon.js'),
            Path('chart-timestack.js'),
            Path('chart-zoom.js'),
            # spin.js is implicit and must be last
        ]
        self.plugin_assets = []  # filled by load_plugins() below

    def set_backend(self, backend):
        self._backend = backend

    def handle_auth(self, request, creds):
        if hdr := request.headers.get(hdrs.AUTHORIZATION):
            try:
                auth = BasicAuth.decode(hdr)
            except ValueError:
                pass
            else:
                cred = (auth.login, sha256(auth.password.encode()).hexdigest())
                if cred in creds:
                    return
        raise web.HTTPUnauthorized(headers={'WWW-Authenticate': 'Basic'})

    def generate_nav(self, cur_page):
        cur_parent = self._conf.pages[cur_page].parent
        nav = ['<nav><ul>']
        # if the current page has a parent, add a "back" link
        if cur_parent:
            parent_conf = self._conf.pages.get(cur_parent)
            if parent_conf and not parent_conf.hidden:
                title = html.escape(parent_conf.title or cur_parent)
                nav.append('<li class="item">'
                           f'<a href="{cur_parent}">&larr; {title}</a></li>')
        children = []
        for page, pageconf in self._conf.pages.items():
            if pageconf.hidden:
                continue
            icon = '<img src="s/locked.svg">' if pageconf.auth else ''
            title = html.escape(pageconf.title or page)
            if page == cur_page:
                nav.append(f'<li class="active item">{icon}{title}</li>')
            else:
                navitem = ('<li class="item">'
                           f'<a href="{page}">{icon}{title}</a></li>')
                # only show pages on the same level
                if pageconf.parent == cur_parent:
                    nav.append(navitem)
                # or children of the current page
                elif pageconf.parent == cur_page:
                    children.append(navitem)
        if children:
            nav.append('<li class="sep">|</li>')
            nav.extend(children)
        nav.append('</ul><ul>'
                   '<li class="clock"><span id="nav-clock"></span></li>'
                   '<li class="logo"><img src="s/mlz.svg"></li>'
                   '</ul></nav>')
        return ''.join(nav)

    async def page(self, request):
        page = request.match_info.get('page', 'default')
        pageconf = self._conf.pages.get(page)
        if not pageconf:
            raise web.HTTPNotFound
        if pageconf.auth:
            self.handle_auth(request, pageconf.auth)
        self.log.info('[%s] serving page %r', request.remote, page)

        # master template
        template = (self._asset_root / 'page.html').read_text()

        # assets
        js = ''.join(f'    <script src="s/{fn}"></script>\n'
                     for fn in self.builtin_assets if fn.suffix == '.js')
        # plugin assets get a numbered "subdirectory" to avoid name clashes
        js += ''.join(f'    <script src="p/{i}/{fn.name}"></script>\n'
                      for i, fn in enumerate(self.plugin_assets)
                      if fn.suffix == '.js')
        jsname = Path(pageconf.file).with_suffix('.js')
        if (self._conf_root / jsname).is_file():
            js += f'    <script src="c/{jsname}"></script>\n'

        cs = ''.join(f'    <link rel="stylesheet" href="s/{fn}">\n'
                     for fn in self.builtin_assets if fn.suffix == '.css')
        cs += ''.join(f'    <link rel="stylesheet" href="p/{i}/{fn.name}">\n'
                      for i, fn in enumerate(self.plugin_assets)
                      if fn.suffix == '.css')
        cssname = Path(pageconf.file).with_suffix('.css')
        if (self._conf_root / cssname).is_file():
            cs += f'<link rel="stylesheet" href="c/{cssname}">'
        template = template.replace('{{JS}}', js).replace('{{CSS}}', cs)

        # main page content
        content = (self._conf_root / pageconf.file).read_text()
        template = template.replace('{{CONTENT}}', content)

        # navigation bar
        show_nav = pageconf.nav_bar
        if show_nav is None:
            show_nav = self._conf.nav_bar and not pageconf.hidden
        if show_nav:
            template = template.replace('{{NAVBAR}}', self.generate_nav(page))
        else:
            template = template.replace('{{NAVBAR}}', '')

        # quick bar
        show_quick = pageconf.quick_icons
        if show_quick is None:
            show_quick = self._conf.quick_icons
        if show_quick:
            template = template.replace('{{QUICKSTYLE}}', '')
        else:
            template = template.replace('{{QUICKSTYLE}}', 'display: none')

        # further template replacements
        if url := self._conf.control_url:
            template = template.replace('{{CONTROL}}', url)
            template = template.replace('{{CONTROL_DISPLAY}}', 'inline')
        else:
            template = template.replace('{{CONTROL}}', '')
            template = template.replace('{{CONTROL_DISPLAY}}', 'none')
        template = template.replace('{{THEME}}', 'dark'
                                    if self._conf.dark_theme else 'light')
        template = template.replace('{{READONLY}}',
                                    str(pageconf.readonly).lower())
        template = template.replace('{{WITHAUTH}}',
                                    str(bool(pageconf.auth)).lower())
        hostname = socket.getfqdn().split('.')[0]
        title = f'{hostname}: {pageconf.title or page}'
        template = template.replace('{{TITLE}}', html.escape(title))

        return web.Response(text=template, content_type='text/html')

    async def static(self, request):
        subpath = sanitize_path(request.match_info['file'])
        if request.match_info['path'] == 's':
            filepath = self._asset_root / subpath
        elif request.match_info['path'] == 'p':
            try:
                filepath = self.plugin_assets[int(subpath.parts[0])]
            except (ValueError, IndexError):
                self.log.warning('[%s] invalid plugin asset name %s',
                                 request.remote, subpath)
                raise web.HTTPNotFound from None
        elif request.match_info['path'] == 'c':
            filepath = self._conf_root / subpath
        else:
            raise web.HTTPNotFound
        self.log.info('[%s] serving %s', request.remote, filepath)
        if filepath.is_file():
            return web.FileResponse(filepath)
        self.log.warning('[%s] %s not found', request.remote, filepath)
        raise web.HTTPNotFound

    async def ws(self, request):
        ws = web.WebSocketResponse()
        self.log.info('[%s] [ws] connection accepted', request.remote)
        await ws.prepare(request)

        try:
            await self._backend.send_all_updates(ws)
        except OSError as err:
            self.log.warning('[%s] [ws] connection error during initial '
                             'update: %s', request.remote, err)
            return

        self.connections.add(ws)
        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                try:
                    req_msg = Request.unserialize(ws, msg.data)
                    self.log.info('got request: %r', req_msg)
                    await self._backend.handle_request(req_msg)
                except Exception:
                    self.log.exception('[%s] [ws] error handling data %r',
                                       request.remote, msg.data)
            elif msg.type == WSMsgType.ERROR:
                self.log.warning('[%s] [ws] connection closed with error %s',
                                 request.remote, ws.exception())

        self.log.info('[%s] [ws] connection closed', request.remote)
        self.connections.discard(ws)
        return ws


def init_app(handler):
    """Set up the aiohttp application and routes."""
    app = web.Application()
    app.router.add_get('/', handler.page, name='default_page')
    app.router.add_get('/ws', handler.ws, name='websocket')
    app.router.add_get(r'/{page:\w+}', handler.page, name='page')
    app.router.add_get(r'/{path:[spc]}/{file:[/\w.-]+}', handler.static,
                       name='static')
    return app


def parse_args(argv):
    parser = argparse.ArgumentParser(description='Start Spin daemon.')
    parser.add_argument('-c', dest='confdir', action='store',
                        type=Path, default=default_cfgdir,
                        help='configuration directory (default '
                        f'{default_cfgdir})')
    parser.add_argument('-v', dest='verbose', action='store_true',
                        help='enable debug logging')
    return parser.parse_args(argv)


def load_plugins(args, conf, loop, handler, backend):
    for path in conf.plugins:
        pluginpath = Path(path)
        try:
            if pluginpath.suffix == '':
                # builtin plugin
                mod = __import__(f'spin.plugins.{path}', None, None, ['*'])
                namespace = mod.__dict__
                pluginpath = Path(__file__).parent / 'plugins' / f'{path}'
            else:
                pluginpath = args.confdir / pluginpath
                namespace = {'__file__': str(pluginpath)}
                # pylint: disable=exec-used
                exec(pluginpath.read_bytes(), namespace)  # noqa: S102
        except Exception as err:  # noqa: BLE001
            mlzlog.log.error('could not load plugin %s: %s', path, err)
            continue
        try:
            res = namespace['init_plugin'](conf, loop)
        except Exception as err:  # noqa: BLE001
            mlzlog.log.error('could not initialize plugin %s: %s', path, err)
            continue
        if 'protos' in res:
            backend.protos.update(res['protos'])
        if 'assets' in res:
            plugindir = pluginpath if pluginpath.is_dir() else pluginpath.parent
            handler.plugin_assets.extend(
                plugindir / a for a in res.get('assets', []))


def main():
    """Run the spin service (backend and HTTP) until interrupted."""
    args = parse_args(sys.argv[1:])
    rootlevel = 'debug' if args.verbose else 'info'
    # log to stdout, as a daemon it goes to journal anyway
    mlzlog.initLogging('spin', rootlevel, logdir=None,
                       tracebacks_on_console=args.verbose)

    conf = get_config(args)
    loop = asyncio.get_event_loop()
    exitstatus = [0]

    handler = SiteHandler(conf, args.confdir, mlzlog.getLogger('handler'))
    backend = Backend(loop, handler.connections, mlzlog.getLogger('backend'))
    handler.set_backend(backend)

    load_plugins(args, conf, loop, handler, backend)

    # handle signals to stop completely
    def stop_everything(sig):
        mlzlog.log.info('stopping due to %s', sig)
        for task in asyncio.all_tasks(loop):
            task.cancel()
    loop.add_signal_handler(signal.SIGINT, stop_everything, 'SIGINT')
    loop.add_signal_handler(signal.SIGTERM, stop_everything, 'SIGTERM')
    loop.set_exception_handler(error_fallback)

    # handle configuration change by a graceful reload on the frontends
    async def stop_and_restart():
        await backend.notify_restart()
        stop_everything('config change')
        exitstatus[0] = 3
    loop.create_task(config_watcher(args, stop_and_restart))  # noqa: RUF006

    def print_to_log(s):
        # in the initial "Running on ..." message, cut the "Ctrl-C" second line
        mlzlog.log.info(s.splitlines()[0])

    backend.start(conf, handler.builtin_assets)
    app = init_app(handler)
    host, port = extract_addr(conf.server, 8000)
    try:
        web.run_app(app, host=host, port=port, loop=loop, handle_signals=False,
                    print=print_to_log)
    except asyncio.CancelledError:
        mlzlog.log.info('server stopped')
    sys.exit(exitstatus[0])



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Various utils."""

import asyncio
import base64
import socket
from collections import defaultdict

import mlzlog

from spin.backend import State


def error_fallback(_loop, context):
    """Fallback exception handler."""
    msg = 'Uncaught exception in Task'
    if (f := context.get('future')) is not None:
        funcname = f.get_coro().__qualname__
        msg += f' "{f.get_name()}" running function {funcname}()'
    mlzlog.log.error(msg, exc_info=context.get('exception'))


def decode_base64_file(b64):
    contents = b64.removeprefix('data:application/octet-stream;base64,')
    return base64.b64decode(contents)


def extract_addr(addr, defaultport=None):
    """Extract host and port from a host:port string."""
    if ':' in addr:
        host, port = addr.split(':', 1)
        if not port.isdigit():
            raise ValueError(f'Invalid port number: {port}')
        return host, int(port)
    if defaultport is None:
        raise ValueError('No port specified and no default port provided')
    return addr, defaultport


class StreamConnection:
    """Base class for line-based protocol connections."""

    default_port = None

    def __init__(self, addr, log):
        self.log = log
        self.addr = addr
        self.devices = defaultdict(list)
        self.conn_write = None
        self.conn_read = None
        self.is_initialized = False

    async def _send(self, msg):
        self.conn_write.write(msg + b'\n')
        await self.conn_write.drain()

    async def _receive(self):
        msg = await self.conn_read.readuntil(b'\n')
        return msg.decode('utf-8')

    async def _open_connection(self):
        """Open the socket connection."""
        host, port = extract_addr(self.addr, self.default_port)
        reader, writer = await asyncio.open_connection(host, port,
                                                       family=socket.AF_INET)
        self.conn_read = reader
        self.conn_write = writer

    async def _synchronize(self):
        raise NotImplementedError

    async def _handle_message(self, msg):
        """Handle incoming messages - protocol specific."""
        raise NotImplementedError

    def add_device(self, device):
        """Add a device to be managed by this connection."""
        raise NotImplementedError

    def _decode(self, line):
        """Protocol-specific line handling."""
        raise NotImplementedError

    async def run(self):
        """Run the main connection loop with retry logic."""
        sleep = 1
        while True:
            self.is_initialized = False
            try:
                await self._open_connection()
                await self._synchronize()
            except Exception as e:
                self.log.exception('initializing connection')
                for devs in self.devices.values():
                    for dev in devs:
                        await dev.send(State.ERROR, f'Error in init: {e}')
                await asyncio.sleep(sleep)
                sleep = min(30, sleep * 1.53)
                continue
            self.is_initialized = True

            self.log.info('handling regular updates')
            while True:
                try:
                    line = await self._receive()
                    await self._handle_message(self._decode(line))
                except Exception as e:
                    self.log.exception('handling message')
                    for devs in self.devices.values():
                        for dev in devs:
                            await dev.send(State.ERROR,
                                           f'Error in connection: {e}')
                    sleep = 1
                    break



# Configuration file for the Sphinx documentation builder.

extensions = []

project = 'Spin'
copyright = '2024- by Georg Brandl, Enrico Faulhaber, Alexander Zaft'  # noqa: A001
author = 'Georg Brandl, Enrico Faulhaber, Alexander Zaft'

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

master_doc = 'index'
default_role = 'literal'
highlight_language = 'python'

html_theme = 'furo'
html_logo = '_static/logo.png'
html_theme_options = {
    'light_css_variables': {
        'font-stack': 'Open Sans, DejaVu Sans, sans-serif',
    },
}
html_static_path = ['_static']



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""PILS backend via Zapf."""

import asyncio
from typing import ClassVar

import zapf.io
from zapf.scan import Scanner
from zapf.spec import DevStatus, ReasonMap, decode_bitfield

from .backend import Reply, Request, State
from .device import PeriodicPollDevice

STATE_MAP = {
    DevStatus.RESET:    State.UNKNOWN,
    DevStatus.IDLE:     State.OK,
    DevStatus.DISABLED: State.DISABLED,
    DevStatus.WARN:     State.WARN,
    DevStatus.START:    State.BUSY,
    DevStatus.BUSY:     State.BUSY,
    DevStatus.STOP:     State.BUSY,
    DevStatus.ERROR:    State.ERROR,
    DevStatus.DIAGNOSTIC_ERROR: State.ERROR,
}


class Device(PeriodicPollDevice):
    """Abstraction for PILS devices, using the Zapf library.

    The library is not async, but has an internal polling thread, so that
    most requests should be "non-blocking enough".

    Device uri can have different forms, depending on the IO protocol.
    To the Zapf connection URI, add `#devname`.
    """

    _cache: ClassVar = {}
    _table = None

    def needed_assets(self):
        if self._devconf.options.get('table'):
            yield 'pils-table.js'

    async def init_poll(self):
        devname = self._uri.fragment
        scheme = self._uri.scheme
        zapfuri = self._uri._replace(scheme=scheme.removeprefix('pils+'),
                                     fragment='').geturl()
        if zapfuri.startswith('sim:/'):
            # fix side effect of urllib.parse without a "hostname"
            zapfuri = 'sim:///' + zapfuri[5:]

        # Scanning can block for a long time, so we run it on a side thread
        # for the first Zapf device.  All other devices reuse the same PlcIO,
        # so they must wait for the "scan done" event to be set.
        if zapfuri not in self._cache:
            def scan():
                return list(Scanner(io, log).get_devices())
            entry = self._cache[zapfuri] = [asyncio.Event(), None]
            log = self.log.getChild('zapf')
            io = zapf.io.PlcIO(zapfuri, log)
            entry[1] = await asyncio.to_thread(scan)
            # This starts the polling thread.
            io.start_cache()
            entry[0].set()
        else:
            await self._cache[zapfuri][0].wait()

        for dev in self._cache[zapfuri][1]:
            if dev.name == devname:
                self._dev = dev
                break
        else:
            raise RuntimeError(f'device {devname} not present in PLC')

        if tblname := self._devconf.options.get('table'):
            try:
                self._table = self._dev.tables[tblname]
            except KeyError:
                raise RuntimeError(f'table {tblname!r} not found in device '
                                   f'{devname!r}') from None

        self._useenum = self._devconf.options.get('enum', False)
        self._auxdef = self._dev.info.get('aux', {})
        self._unit = self._dev.info['unit']
        self._params = {}
        for i, param in enumerate(self._devconf.params):
            if param.name == 'target':
                self._params['target'] = (self._unit, param.poll_every,
                                          i % param.poll_every)
                continue

            try:
                unit = self._dev.get_param_valueinfo(param.name).unit
            except zapf.Error as e:
                self.log.error('could not find info for parameter %s, '
                               'ignoring it', param.name)
                await self.send_param(
                    param.name, None, '', f'Could not query unit: {e}')
            else:
                self._params[param.name] = (unit, param.poll_every,
                                            i % param.poll_every)

    async def poll(self, n):
        try:
            state, reason, aux, _ = self._dev.read_status()
        except zapf.Error as e:
            await self.send(State.ERROR, f'Error in status poll: {e}')
            return

        state = STATE_MAP.get(state, State.UNKNOWN)
        string_list = []
        reason = ReasonMap[reason]
        if reason:
            string_list.append(reason)
        if aux:
            string_list.append(decode_bitfield(aux, self._auxdef))
        status = ', '.join(string_list)

        try:
            if self._useenum:
                val = self._dev.read_value()
            else:
                val = self._dev.read_value_raw()
        except zapf.Error as e:
            await self.send(State.ERROR, f'Error in value poll: {e}')
        else:
            await self.send(state, status, val)

        for param, (unit, every, stagger) in self._params.items():
            if n == 0 or n % every == stagger:
                try:
                    if param == 'target':
                        if self._useenum:
                            res = self._dev.read_target()
                        else:
                            res = self._dev.read_target_raw()
                    else:
                        # get_param() can block for a long time due to
                        # manipulation of the paramctrl interface
                        # TODO: avoid for FlatParam devices
                        res = await asyncio.to_thread(
                            lambda p=param: self._dev.get_param(p))
                except zapf.Error as e:
                    await self.send_param(param, None, '',
                                          f'Error in param poll: {e}')
                else:
                    await self.send_param(param, res, unit)

    def _value_type_from_info(self, info):
        if info.basetype == 'float':
            return ['float', info.min_value, info.max_value]
        if info.basetype in ('int', 'uint'):
            return ['int', info.min_value, info.max_value]
        if info.basetype == 'enum':
            return ['choice', list(info.enum_w)]
        return ['float', -1e308, 1e308]

    async def input_value_type(self, param):
        try:
            if param == 'value':
                info = self._dev.value_info
            else:
                info = self._dev.get_param_valueinfo(param)
        except zapf.Error:
            return ['float', -1e308, 1e308]
        return self._value_type_from_info(info)

    async def handle_request(self, request: Request):
        if request.action == 'read-table':
            return await self._read_table(request)
        if request.action == 'write-cell':
            return await self._write_table_cell(request)
        return await super().handle_request(request)

    async def action_state(self, param):
        if param == 'value':
            return STATE_MAP.get(self._dev.read_status()[0], State.UNKNOWN)
        return State.OK

    async def action_stop(self):
        self._dev.change_status((DevStatus.BUSY,), DevStatus.STOP)

    async def action_reset(self):
        self._dev.reset()

    async def action_get(self, param):
        if param == 'value':
            return self._dev.read_value_raw()
        return self._dev.get_param(param)

    async def action_set(self, param, value):
        if param in {'value', 'target'}:
            if self._useenum:
                self._dev.change_target(value)
            else:
                self._dev.change_target_raw(value)
        else:
            self._dev.set_param(param, value)

    async def action_run(self, command):
        await asyncio.to_thread(lambda: self._dev.exec_func(command))

    async def action_upload(self, filename, file, target=None):
        pass

    async def _read_table(self, request):
        """Read the current table data."""
        if not self._table:
            raise RuntimeError('no table configured for this device')
        rows = self._table.get_size()[0]
        colnames = self._table.list_columns()
        cols = []
        for name in colnames:
            value_info = self._table.get_column_valueinfo(name)
            if value_info.readonly:
                cols.append((name, None))
            else:
                cols.append((name, self._value_type_from_info(value_info)))
        data = []
        for i in range(rows):
            await self.send_progress(request.connection, i/rows,
                                     'Reading table...')
            row_data = []
            for col in colnames:
                try:
                    value = self._table.get_cell(i, col)
                except zapf.Error as e:
                    raise RuntimeError('Error reading table cell '
                                       f'{i}, {col}: {e}') from None
                row_data.append(value)
            data.append(row_data)
        await self.send_progress(request.connection, 1)
        return Reply(request.connection, request.key, None,
                     'pilstable', {
                         'rows': rows,
                         'cols': cols,
                         'data': data,
                     })

    async def _write_table_cell(self, request):
        """Write a single cell in the table."""
        row, col = request.data['row'], request.data['col']
        self._table.set_cell(row, col, request.data['value'])
        new_value = self._table.get_cell(row, col)
        return Reply(request.connection, request.key, None,
                     'pilscell', {
                         'row': row,
                         'col': col,
                         'value': new_value,
                     })



# pylint: skip-file

import asyncio
import random

from spin.backend import Reply, Request, State
from spin.device import BaseDevice


class PluginDevice(BaseDevice):
    """A demo plugin device for testing purposes."""

    def init(self):
        self.run_task(self._poll())

    async def _poll(self):
        self.log.info('plugin device polling')
        while True:
            await self.send(State.OK, 'ok', random.random())
            await asyncio.sleep(1)

    async def handle_request(self, request: Request):
        if request.action == 'click' and request.intent == 'set':
            self.run_task(self._update(request.connection))
            return Reply(request.connection, request.key, request.transform,
                         'demoplugin-notify', 'working...')
        return None

    async def _update(self, conn):
        self.log.info('starting update')
        for i in range(101):
            label = 'Herding llamas...'
            if i > 30:
                label = 'Reticulating splines...'
            if i > 70:
                label = 'Almost there...'
            await self.send_progress(conn, i/100, label)
            await asyncio.sleep(random.random() / 10)  # Simulate some work
        self.log.info('update done')
        await self.send_reply(Reply(conn, self._name, None,
                                    'demoplugin-notify', 'success!'))


def init_plugin(_conf, _loop):
    return {
        'protos': {'demoplugin': PluginDevice},
        'assets': ['plugin-notify.js', 'plugin.css'],
    }



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Tango backend."""

import ast

import tango
from tango import DevFailed, DevState
from tango.asyncio import DeviceProxy as AsyncDeviceProxy

from .backend import State
from .device import PeriodicPollDevice

STATE_MAP = {
    DevState.ALARM: State.WARN,
    DevState.CLOSE: State.DISABLED,
    DevState.DISABLE: State.DISABLED,
    DevState.FAULT: State.ERROR,
    DevState.INIT: State.BUSY,
    DevState.MOVING: State.BUSY,
    DevState.OFF: State.DISABLED,
    DevState.ON: State.OK,
    DevState.OPEN: State.OK,
    DevState.RUNNING: State.BUSY,
    DevState.STANDBY: State.DISABLED,
}

NUMERIC_TYPES = {
    tango.DevUChar: ['int', 0, 255],
    tango.DevShort: ['int', -32768, 32767],
    tango.DevUShort: ['int', 0, 65535],
    tango.DevLong: ['int', -2**31, 2**31 - 1],
    tango.DevULong: ['int', 0, 2**32 - 1],
    tango.DevLong64: ['int', -2**63, 2**63 - 1],
    tango.DevULong64: ['int', 0, 2**64 - 1],
    tango.DevFloat: ['float', -3e38, 3e38],
    tango.DevDouble: ['float', -1e308, 1e308],
}


def format_fail(exc):
    """Return a *simple* one-line exception string from a convoluted tango exc.

    Takes the first line of the exception's string representation that looks
    like it could be the topmost "description" of the exception.
    """
    lines = str(exc).splitlines()
    firstline = ''
    desc = ''
    for raw_line in lines:
        line = raw_line.strip()
        if line.endswith(('DevError[', 'DevFailed[')):
            continue
        if not firstline:
            firstline = line
        if line.startswith('desc ='):
            desc = line[6:].strip()
        if line.startswith('reason =') and desc:
            desc = f'{line[8:].strip()}: {desc}'
            break
    return desc or firstline


def parse_int_mapping(mapping):
    """Parse the "mapping" property of digital devices."""
    labels = []
    val2label = {}
    label2val = {}
    mapping = ast.literal_eval(mapping)
    for entry in mapping:
        parts = entry.split(':')
        if len(parts) < 2:
            continue
        try:
            val = int(parts[0].strip())
        except ValueError:
            continue
        label = parts[1].strip()
        val2label[val] = label
        label2val[label] = val
        if len(parts) == 2 or 'ro' not in parts[2]:
            labels.append(label)
    return labels, val2label, label2val


class Device(PeriodicPollDevice):
    """Abstraction for Tango devices, using the PyTango asyncio integration.

    Device url should look like: 'tango://<host>:<port>/<x>/<y>/<z>',
    with x/y/z being the device name.
    """

    _dev = None
    _params = None

    # TODO: this assumes MLZ style devices with 'value' attribute

    async def init_poll(self):
        self._dev = await AsyncDeviceProxy(self._devconf.device)
        self._attrconf = {}
        self._props = {}
        self._mapping = None, None, None

        for elem in self._dev.attribute_list_query():
            self._attrconf[elem.name] = elem
        unit = self._attrconf['value'].unit

        try:
            proplist = await self._dev.command_inout('GetProperties')
        except DevFailed as e:
            self.log.warning('could not query properties: %s', format_fail(e))
        else:
            for i in range(0, len(proplist), 2):
                self._props[proplist[i]] = proplist[i + 1]

            if self._devconf.options.get('enum') and 'mapping' in self._props:
                self._mapping = parse_int_mapping(self._props['mapping'])

        self._unit = unit if unit != 'No unit' else ''
        self._params = {}
        self._send_target = False

        for i, param in enumerate(self._devconf.params):
            if param.name == 'target':
                self._send_target = True
                continue

            try:
                unit = self._attrconf[param.name].unit
            except DevFailed as e:
                self.log.error('could not find info for parameter %s, '
                               'ignoring it', param.name)
                await self.send_param(
                    param.name, None, '',
                    f'Could not query unit: {format_fail(e)}')
            else:
                unit = unit if unit != 'No unit' else ''
                self._params[param.name] = (unit, param.poll_every,
                                            i % param.poll_every)

    async def poll(self, n):
        try:
            state = await self._dev.State()
            status = await self._dev.Status()
        except DevFailed as e:
            await self.send(State.ERROR,
                            f'Error in status poll: {format_fail(e)}')
            return

        state = STATE_MAP.get(state, State.UNKNOWN)
        if state in (State.ERROR, State.UNKNOWN):
            await self.send(state, status)
            return

        try:
            val_attr = await self._dev.read_attribute('value')
            read_val = val_attr.value
            write_val = val_attr.w_value
        except DevFailed as e:
            await self.send(State.ERROR,
                            f'Error in value poll: {format_fail(e)}')
        else:
            if self._mapping[1]:
                read_val = self._mapping[1].get(read_val, str(read_val))
            await self.send(state, status, read_val)
            if self._send_target:
                if self._mapping[2]:
                    write_val = self._mapping[2].get(write_val, str(write_val))
                await self.send_param('target', val_attr.w_value, self._unit)

        for param, (unit, every, stagger) in self._params.items():
            if n == 0 or n % every == stagger:
                try:
                    res = await self._dev.read_attribute(param)
                except DevFailed as e:
                    await self.send_param(
                        param, None, '',
                        f'Error in param poll: {format_fail(e)}')
                else:
                    await self.send_param(param, res.value, unit)

    def format_exception(self, exc):
        if isinstance(exc, DevFailed):
            return format_fail(exc)
        return str(exc)

    async def input_value_type(self, param):
        try:
            dtype = self._attrconf[param].data_type
        except KeyError:
            return ['float', -1e308, 1e308]

        absmin = absmax = 0
        if param == 'value':
            if self._mapping[0]:
                return ['choice', self._mapping[0]]
            absmin = self._props.get('absmin', 0)
            absmax = self._props.get('absmax', 0)

        if dtype in NUMERIC_TYPES:
            if absmin == absmax == 0:
                absmin = NUMERIC_TYPES[dtype][1]
                absmax = NUMERIC_TYPES[dtype][2]
            return [NUMERIC_TYPES[dtype][0], absmin, absmax]
        return ['str']

    async def action_state(self, param):
        if param == 'value':
            return STATE_MAP.get(await self._dev.State(), State.UNKNOWN)
        return State.OK

    async def action_stop(self):
        await self._dev.Stop()

    async def action_reset(self):
        await self._dev.Reset()

    async def action_get(self, param):
        return (await self._dev.read_attribute(param)).value

    async def action_set(self, param, value):
        if param == 'value' and self._mapping[2] and value in self._mapping[2]:
            value = self._mapping[2][value]
        if param == 'target':
            param = 'value'
        await self._dev.write_attribute(param, value)

    async def action_run(self, command):
        await self._dev.command_inout(command)

    async def action_upload(self, filename, file, target=None):
        pass



# *****************************************************************************
# Nobody expects the Spanish Inquisition!
# Copyright (c) 2024- by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Alexander Zaft <a.zaft@fz-juelich.de>
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Base implementations for backend devices."""

import asyncio
from contextlib import asynccontextmanager
from typing import ClassVar

from .backend import Reply, Request, State, Update
from .utils import decode_base64_file


class BaseDevice:
    """Base abstraction for a single device."""

    def __init__(self, name, devconf, uri, backend, log):
        self.log = log
        self._uri = uri
        self._name = name
        self._devconf = devconf
        self._backend = backend
        self._unit = ''

    @property
    def is_initialized(self):
        """Can return False if the device is not yet initialized."""
        return True

    # Utility methods.

    def run_task(self, coro):
        self._backend.run_task(coro)

    async def send(self, state, state_str, value=None):
        """Call this to send a main value update to all frontends."""
        await self._backend.send_update(Update(
            self._name,
            self._unit,
            state,
            state_str,
            value,
        ))

    async def send_param(self, name, value, unit, error=None):
        """Call this to send a parameter update to all frontends."""
        await self._backend.send_update(Update(
            self._name + '.' + name,
            unit,
            State.OK if error is None else State.ERROR,
            error or '',
            value,
        ))

    async def send_reply(self, reply: Reply):
        await self._backend.send_reply(reply)

    async def send_progress(self, conn, progress, label=''):
        """Send a progress update to the given frontend.

        *progress* should be a float between 0 and 1. *label* is not required
        if progress is 1 (done).
        """
        await self.send_reply(Reply(conn, self._name, None, 'progress',
                                    [label, progress]))

    @asynccontextmanager
    async def error_guard(self, conn, msg=None, *, reset_progress=True):
        """Catch errors of long running tasks spawned for this device.

        The error message will be appended to 'msg'.
        """
        try:
            yield
        except Exception as err:
            if reset_progress:
                await self.send_progress(conn, 1)
            msg = (msg or 'Error in task: ') + str(err)
            self.log.exception(msg)
            await self.send_reply(Reply(conn, self._name, None, 'error', msg))

    # Overridable methods.

    def needed_assets(self):
        """Return a list of CSS/JS assets needed by this device."""
        return []

    def init(self):
        raise NotImplementedError

    def format_exception(self, exc):
        """Format exceptions for sending to the frontend.

        This can be overridden by subclasses to avoid overly complex
        information.
        """
        return str(exc)

    async def handle_request(self, request: Request):
        """Handle a Request from a frontend."""
        param = request.key.partition('.')[2] or 'value'

        if not self.is_initialized:
            raise RuntimeError('device not initialized')

        transform = None
        firstbit = None
        bitmask = None
        factor = 1.0
        if request.transform:
            if request.transform[0] == 'bits':
                transform = 'bits'
                firstbit = int(request.transform[1])
                bitmask = (1 << int(request.transform[2])) - 1
            elif request.transform[0] == 'scale':
                transform = 'scale'
                factor = float(request.transform[1])
            else:
                raise ValueError('unknown transform specified')

        # 'click' needs to be confirmed if the device is configured for that
        if request.action == 'click' and request.intent != 'input' and \
           self._devconf.confirm:
            return Reply(request.connection, request.key,
                         request.transform, 'confirm',
                         {'intent': request.intent,
                          'data': request.data,
                          'what': f'{request.intent} on {self._name}',
                          'prompt': self._devconf.confirm})

        # otherwise, 'confirm' and 'click' are the same
        if request.action in ('click', 'confirm'):
            cur_state = await self.action_state(param)

            if request.intent not in ('reset', 'stop', 'input',
                                      'toggle', 'set', 'run'):
                raise ValueError(f'Unknown intent {request.intent}')

            if request.intent == 'reset' or \
               (param == 'value' and cur_state == State.ERROR):
                await self.action_reset()
                return None

            if request.intent == 'stop' or \
               (param == 'value' and cur_state == State.BUSY):
                await self.action_stop()
                return None

            if request.intent == 'run':
                await self.action_run(request.data)
                return None

            if request.intent == 'input':
                valuetype = await self.input_value_type(param)
                return Reply(request.connection, request.key,
                             request.transform, 'input', valuetype)

            new_value = None

            if request.intent == 'toggle':
                cur_value = await self.action_get(param)
                if transform == 'bits':
                    field_value = (cur_value >> firstbit) & bitmask
                    new_value = cur_value & ~(bitmask << firstbit)
                    new_value |= int(not field_value) << firstbit
                else:
                    toggle_vals = request.data or [0, 1]
                    new_value = toggle_vals[1] if cur_value == toggle_vals[0] \
                        else toggle_vals[0]

            elif request.intent == 'set':
                if transform == 'bits':
                    cur_value = await self.action_get(param)
                    new_value = cur_value & ~(bitmask << firstbit)
                    new_value |= (request.data & bitmask) << firstbit
                elif transform == 'scale':
                    new_value = request.data / factor
                else:
                    new_value = request.data

            await self.action_set(param, new_value)

        elif request.action == 'input':
            # got a specific value to set, treat it like a click
            request.action = 'click'
            request.intent = 'set'
            return await self.handle_request(request)

        elif request.action == 'upload':
            decoded = decode_base64_file(request.data['contents'])
            await self.action_upload(request.data['filename'], decoded)
            return None

        else:
            raise ValueError(f'Unknown action {request.action}')

        # after an action, make sure we poll the new state directly
        if isinstance(self, PeriodicPollDevice):
            await self.poll(0)
        return None

    # The following methods are used by the default handle_request() and
    # can be ignored by subclasses that define their own.

    async def action_state(self, param):
        """Get the state of the device or a parameter, for actions."""
        raise NotImplementedError

    async def action_stop(self):
        """Stop the device, for actions."""
        raise NotImplementedError

    async def action_reset(self):
        """Reset the device, for actions."""
        raise NotImplementedError

    async def action_get(self, param):
        """Get the value of a parameter, for actions."""
        raise NotImplementedError

    async def action_set(self, param, value):
        """Set the value of a parameter, for actions."""
        raise NotImplementedError

    async def action_run(self, command):
        """Execute a backend-specific command."""
        raise NotImplementedError

    async def action_upload(self, filename, file, target=None):
        """Upload a file to the device.

        Select which action to do by specifying target.
        """
        raise NotImplementedError

    async def input_value_type(self, _param):
        """Return the type of a parameter, for querying input.

        Possible return values are:

        - ['int', min, max]
        - ['float', min, max]
        - ['str']
        - ['choice', ['value1', 'value2', ...]]
        """
        return ['float', -1e308, 1e308]


class PeriodicPollDevice(BaseDevice):
    """Base class for devices which periodically poll their endpoint."""

    _is_initialized = False

    def init(self):
        self.run_task(self.periodic_poll())

    @property
    def is_initialized(self):
        return self._is_initialized

    async def periodic_poll(self):
        sleep = 1
        while True:
            n = 0
            self._is_initialized = False
            try:
                await self.init_poll()
            except Exception as e:
                self.log.exception('error initializing poller')
                exc_str = self.format_exception(e)
                await self.send(State.ERROR, f'Could not init: {exc_str}')
                await asyncio.sleep(sleep)
                sleep = min(30, sleep * 1.53)
                continue
            self.log.info('poller initialized')
            self._is_initialized = True

            while True:
                try:
                    await self.poll(n)
                    n += 1
                except Exception as e:
                    self.log.exception('error in poll')
                    exc_str = self.format_exception(e)
                    await self.send(State.ERROR, f'Error in poll: {exc_str}')

                await asyncio.sleep(self._devconf.pollinterval)

    async def init_poll(self):
        raise NotImplementedError

    async def poll(self, n):
        raise NotImplementedError


class AsyncUpdateDevice(BaseDevice):
    # pylint: disable=abstract-method
    """Base class for devices that support async updates.

    This provides common functionality for devices that maintain
    a cache of status/value and handle parameter updates.
    """

    _cache: ClassVar = {}
    # must be something that has an `add_device` method, possibly
    # a spin.utils.StreamConnection
    _connclass = object

    def __init__(self, name, devconf, uri, backend, log):
        super().__init__(name, devconf, uri, backend, log)
        self.cache = {
            'status': (State.UNKNOWN, ''),
            'value': None,
        }
        self.extra_params = [par.name for par in devconf.params]
        self._addr, self.device = uri.netloc, uri.path.lstrip('/')

    def init(self):
        if self._addr not in self._cache:
            self._cache[self._addr] = self._connclass(
                self._addr, self.log.parent.getChild(self._addr))
            self.run_task(self._cache[self._addr].run())
        self._conn = self._cache[self._addr]
        self._conn.add_device(self)

    @property
    def is_initialized(self):
        return self._conn.is_initialized

    async def action_state(self, param):
        if param == 'value':
            return self.cache['status'][0]
        return State.OK

    async def action_get(self, param):
        return self.cache[param]



# This is a Zapf simulated PLC that runs automatically for the demo config.

# pylint: skip-file
# ruff: noqa: PGH004
# ruff: noqa

import time

from zapf.simulator.funcs import adr, memcpy, memset, sizeof
from zapf.simulator.runtime import (
    Globals,
    Struct,
    Var,
    array,
    byte,
    dword,
    program,
    real,
    string,
    word,
)
from zapf.spec import FLOAT32_MAX
from zapf.spec.v_2021_09 import DEVICE_FLAGS, encode_unit

PILS_MAGIC = 2021.09
DESCRIPTOR_SLOT_SIZE = 48
DESCRIPTOR_SLOTS = 4
INDEXER_DATA_SIZE = DESCRIPTOR_SLOTS * DESCRIPTOR_SLOT_SIZE
INDEXER_SIZE = 2 + INDEXER_DATA_SIZE
INDEXER_OFFSET = 6
PARAM_DEFAULTS = [0, 100, 10, 90, 5, 500, 0, 10, 2, 0, 0.125, 0, 10, 1, 0, 0, 0, 0]


class DevicesLayout(Struct):
    # di32, do32  (re-using do32_target as di32_value)
    dx32_value = Var(dword, 0)
    dx32_target = Var(dword, 0)
    dx32_estatus = Var(dword, 0x10000000)
    dx32_padding = Var(dword, 0)

    # pi/po32
    fax32_params = Var(array(real, 0, 17), PARAM_DEFAULTS)
    px32_value = Var(real, 50)
    px32_target = Var(real, 0)
    px32_estatus = Var(dword, 0x10000000)
    px32_nerrid = Var(word, 0)
    px32_pctl = Var(word, 0)
    px32_pvalue = Var(real, 0)

    # table storage: 1d
    table32_act_row = Var(word, 0)
    table32_req_row = Var(word, 0)
    table32_line = Var(array(real, 0, 9))
    # table storage: 2d, direct mapped
    table32 = Var(array(real, 0, 99))


alignment_ok = True
for _n, addr in DevicesLayout.OFFSET.items():
    o = getattr(DevicesLayout, _n)
    try:
        size = o.dtype.INNER.sizeof()
    except AttributeError:
        size = o.dtype.sizeof()
    if addr != (addr // size) * size:
        print(f'{_n} is not correctly aligned! (addr={addr}, size={size})')
        alignment_ok = False
assert alignment_ok


def addrof(n):
    return 200 + DevicesLayout.OFFSET[n]


class PLCDescriptor(Struct):
    dt = Var(word, 0x1010)
    devices = Var(word, 0)
    description = Var(word, 0)
    version = Var(word, 0)
    author = Var(word, 0)
    desc_slot_size = Var(word, DESCRIPTOR_SLOT_SIZE)
    num_devices = Var(word, 0)
    flags = Var(word, (DESCRIPTOR_SLOTS - 1) << 12)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 16 - 1), '')
    # flags, desc_slot_size == const, num_devices -> magic!
    keys = 'dt', 'devices', 'description', 'version', 'author', 'name'


class DeviceDescriptor(Struct):
    dt = Var(word, 0x2014)
    prev = Var(word, 0)
    description = Var(word, 0)
    value_param = Var(word, 0)
    aux = Var(word, 0)
    params = Var(word, 0)
    errid = Var(word, 0)
    typecode = Var(word, 0)
    address = Var(word, 0)
    flags = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 20 - 1), '')
    keys = 'dt', 'prev', 'description', 'value_param', 'aux', 'params', \
        'errid', 'typecode', 'address', 'flags', 'name'


class StringDescriptor(Struct):
    dt = Var(word, 0x3004)
    prev = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 4 - 1), '')
    keys = 'dt', 'prev', 'name'


class EnumDescriptor(Struct):
    dt = Var(word, 0x4006)
    prev = Var(word, 0)
    # don't name a field 'value' or 'addr' or the simulator may ignore writes to it...
    val = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 6 - 1), '')
    keys = 'dt', 'prev', 'val', 'name'


class BitfieldDescriptor(Struct):
    dt = Var(word, 0x5008)
    prev = Var(word, 0)
    enum_id = Var(word, 0)
    lsb = Var(byte, 0)
    width = Var(byte, 1)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 8 - 1), '')
    keys = 'dt', 'prev', 'enum_id', 'lsb', 'width', 'name'


class FlagDescriptor(Struct):
    dt = Var(word, 0x5105)
    prev = Var(word, 0)
    lsb = Var(byte, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 5 - 1), '')
    keys = 'dt', 'prev', 'lsb', 'name'


class NumericParameterDescriptor(Struct):
    dt = Var(word, 0x6114)
    prev = Var(word, 0)
    description = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    unit = Var(word, 0)
    minval = Var(real, -FLOAT32_MAX)
    maxval = Var(real, FLOAT32_MAX)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 20 - 1), '')
    keys = 'dt', 'prev', 'description', 'paramidx', 'paramtype', 'unit', \
        'minval', 'maxval', 'name'


class EnumParameterDescriptor(Struct):
    dt = Var(word, 0x620E)
    prev = Var(word, 0)
    description = Var(word, 0)
    read_enum_id = Var(word, 0)
    write_enum_id = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 14 - 1), '')
    keys = 'dt', 'prev', 'description', 'read_enum_id', 'write_enum_id', \
        'paramidx', 'paramtype', 'name'


class SpecialFuncDescriptor(Struct):
    dt = Var(word, 0x680E)
    prev = Var(word, 0)
    description = Var(word, 0)
    arg_id = Var(word, 0)
    res_id = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 14 - 1), '')
    keys = 'dt', 'prev', 'description', 'arg_id', 'res_id', 'paramidx', \
        'paramtype', 'name'


class TableDescriptor(Struct):
    dt = Var(word, 0x6C10)
    prev = Var(word, 0)
    description = Var(word, 0)
    columns = Var(word, 0)
    base = Var(word, 0)
    flags = Var(word, 0)
    last_row = Var(word, 0)
    last_column = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 16 - 1), '')
    keys = 'dt', 'prev', 'description', 'columns', 'base', 'flags', \
        'last_row', 'last_column', 'name'


class DebugDescriptor(Struct):
    dt = Var(word, 0xFF06)
    cycle = Var(word, 0)
    indexer_size = Var(word, 0)
    text = Var(string(INDEXER_DATA_SIZE - 6 - 1), '')  # Yes: this is BIG
    keys = 'dt', 'cycle', 'text'


class ST_Indexer(Struct):
    Request = Var(word, 0)
    #    Data = Var(array(U_Descriptor, 0, DESCRIPTOR_SLOTS-1))
    Data = Var(array(array(byte, 0, DESCRIPTOR_SLOT_SIZE - 1), 0, DESCRIPTOR_SLOTS - 1))


# singletons


PLC_DESCRIPTOR = PLCDescriptor(name='testplc_2021_09.py').alloc_self()
DEBUG_DESCRIPTOR = DebugDescriptor(text='PLC_Problem 42, please call 137',
                                   indexer_size=INDEXER_DATA_SIZE).alloc_self()


# map descriptorid/descriptortuple to descriptor
DESCRIPTOR_ARRAY = [PLC_DESCRIPTOR]
descriptors = {}


def add_descr(t, Descr):
    # pylint: disable=global-statement
    if t in descriptors:
        return descriptors[t]
    next_descriptor = len(DESCRIPTOR_ARRAY)
    kwds = dict(zip(Descr.keys, t))
    d = Descr(**kwds).alloc_self()
    descriptors[t] = next_descriptor
    DESCRIPTOR_ARRAY.append(d)
    return next_descriptor


def add_Device(name, typecode, address, description='', value_param=None,
               aux=0, params=0, errid=0, flags=0, unit=''):
    if value_param is None:
        # provide some sensible default for most devices
        if typecode in (0x1201, 0x1401, 0x1602, 0x1A02, 0x1E03, 0x1E04):
            value_param = add_IntParam(0, '', 'rw', 0, -32768, 32767)
        elif typecode in (0x1202, 0x1402, 0x1604, 0x1A04, 0x1E06):
            value_param = add_IntParam(0, '', 'rw', 0, -(2 ** 31), 2 ** 31 - 1)
        elif typecode in (0x1204, 0x1404, 0x1608, 0x1A08, 0x1E0C):
            value_param = add_IntParam(0, '', 'rw', 0, -(2 ** 63), 2 ** 63 - 1)
        elif typecode >> 8 != 5:  # not for MSGIO !
            value_param = add_FloatParam(0, '', 'rw', 0, -FLOAT32_MAX, FLOAT32_MAX,
                                         encode_unit(unit) if isinstance(unit, str) else unit)
        else:
            value_param = 0
    if isinstance(flags, str):
        flags = [flags]
    try:
        int(flags)
    except TypeError:
        f = 0
        for bit, fn in DEVICE_FLAGS.items():
            if fn in flags:
                fn |= 1 << bit
        flags = f
    t = (
        0x2014,
        PLC_DESCRIPTOR.devices,
        add_String(description),
        value_param,
        aux,
        params,
        errid,
        typecode,
        address,
        flags,
        name,
    )
    PLC_DESCRIPTOR.devices = add_descr(t, DeviceDescriptor)
    PLC_DESCRIPTOR.num_devices += 1
    # no return....


def add_String(s):
    prev = 0
    while s:
        part, s = s[: DESCRIPTOR_SLOT_SIZE - 4 - 1], s[DESCRIPTOR_SLOT_SIZE - 4 - 1:]
        t = (0x3004, prev, part)
        prev = add_descr(t, StringDescriptor)
    return prev


def add_Enum(prev, val, name):
    t = (0x4006, prev, val, name)
    return add_descr(t, EnumDescriptor)


def add_Enums(prev, enumdict):
    # add all values of a dict to the given chain.
    # XXX: also remove duplicates aready in prev? How?
    for v, n in sorted(enumdict.items()):
        prev = add_Enum(prev, v, n)
    return prev


def add_Bitfield(prev, name, lsb, width, enum):
    if isinstance(enum, dict):
        enum = add_Enums(0, enum)
    t = (0x5008, prev, enum, lsb, width, name)
    return add_descr(t, BitfieldDescriptor)


def add_Flag(prev, lsb, name):
    t = (0x5105, prev, lsb, name)
    return add_descr(t, FlagDescriptor)


def add_BfFChain(prev, chain):
    # chain is an iterable of 2/4 tuples for flags/bitfields
    # flags: (bit, name)
    # bitfield: (name, lsb, width, enumdict)
    for e in chain:
        if len(e) == 2:
            prev = add_Flag(prev, *e)
        elif len(e) == 4:
            prev = add_Bitfield(prev, *e)
        else:
            raise RuntimeError(f'bad Element {e} in Bitfield/Flag Chain: {chain}')
    return prev


WRO_MAP = {'rw': 1, 'ro': 2, 'obs': 3}  # writeable, readonly, observable


def add_IntParam(prev, name, wro, idx, minval, maxval, unit=0, description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if isinstance(unit, str):
        unit = encode_unit(unit)
    t = (0x6114, prev, description, idx, (0xC + wro) << 12, unit, minval, maxval, name)
    return add_descr(t, NumericParameterDescriptor)


def add_FloatParam(prev, name, wro, idx, minval, maxval, unit='', description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if isinstance(unit, str):
        unit = encode_unit(unit)
    t = (0x6114, prev, description, idx, (0x4 + wro) << 12, unit, minval, maxval, name)
    return add_descr(t, NumericParameterDescriptor)


def add_EnumParam(prev, name, wro, idx, read_enum, write_enum=0, description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if read_enum == 0:
        read_enum = write_enum
    if isinstance(read_enum, dict):
        if isinstance(write_enum, dict):
            read_enum.update(write_enum)
        read_enum = add_Enums(0, read_enum)
    if isinstance(write_enum, dict):
        write_enum = add_Enums(0, write_enum)
    t = (0x620E, prev, description, read_enum, write_enum, idx, (0x8 + wro) << 12, name)
    return add_descr(t, EnumParameterDescriptor)


def add_SFunc(prev, name, idx, arg_id, res_id, description=''):
    if isinstance(description, str):
        description = add_String(description)
    t = (0x680E, prev, description, arg_id, res_id, idx, 0, name)
    return add_descr(t, SpecialFuncDescriptor)


def add_Table(prev, name, base, typ, columns, last_row, last_column, description=''):
    if isinstance(description, str):
        description = add_String(description)
    t = (0x6C10, prev, description, columns, base, (typ & 3) << 14, last_row, last_column, name)
    return add_descr(t, TableDescriptor)


# prepare some chains to be used later

AUX8 = add_BfFChain(0, [(d, f'AUX{d}') for d in range(6)] +
                    [('Minion', 6, 2, {1: 'Ape', 2: 'Banana', 3: 'Dragon'})])

AUX24 = add_BfFChain(AUX8, [(d, f'AUX{d}') for d in range(8, 20)] +
                     [('Minion', 6, 3, {4: 'Hamster', 2:''})] +
                     [(d, f'AUX{d}') for d in range(20, 24)] +
                     [('', 21, 1, {1:'aux22', 0:'nix'})]+
                     [('X', 18, 3, {})] +
                     [('', 22, 2, {0:'', 1:'Under voltage', 2:'Over voltage', 3:'no Power'})])

ERRID = add_BfFChain(
    0, [('lowerByte', 0, 8, 0), ('topmost 7 Bit', 9, 7, 0), (8, 'a Flag')]
)

PARS_1 = add_FloatParam(0, 'UserMin', 'rw', 0, 0, 100, '#', 'lower user settable limit')
PARS_2 = add_FloatParam(PARS_1, 'UserMax', 'rw', 1, 0, 100, '#', 'upper user settable limit')
PARS_3 = add_FloatParam(PARS_2, 'WarnMin', 'rw', 2, 0, 100, '#', 'lower warn limit')
PARS_4 = add_FloatParam(PARS_3, 'WarnMax', 'rw', 3, 0, 100, '#', 'upper warn limit')
PARS_5 = add_FloatParam(PARS_4, 'Timeout', 'rw', 4, 0, 900, 's', 'timeout for movement in s')
PARS_6 = add_FloatParam(PARS_5, 'MaxTravelDist', 'rw', 5, 0, 500, '#', 'maximum travel distance')
PARS_7 = add_FloatParam(PARS_6, 'Offset', 'ro', 6, 0, 100, '#', 'internal offset')
PARS_8 = add_FloatParam(PARS_7, 'P', 'rw', 7, 0, 100, '%/(#)', 'P constant for regulation')
PARS_9 = add_FloatParam(PARS_8, 'I', 'rw', 8, 0, 100, 's', 'I constant for regulation')
PARS_10 = add_FloatParam(PARS_9, 'D', 'rw', 9, 0, 100, '1/s', 'D constant for regulation')
PARS_11 = add_FloatParam(PARS_10, 'Hysteresis', 'rw', 10, 0, 100, '#', 'Hysteresis for regulation')
PARS_12 = add_FloatParam(PARS_11, 'Holdback', 'rw', 11, 0, 100, '#', 'max difference between actual temp and setpoint')
PARS_13 = add_FloatParam(PARS_12, 'Speed', 'rw', 12, 0.1, 100, '#/s', 'max speed of movement')
PARS_14 = add_FloatParam(PARS_13, 'Accel', 'rw', 13, 0, 100, '#/s^2', 'acceleration of movement')
PARS_15 = add_FloatParam(PARS_14, 'Home', 'rw', 14, 0, 100, '#', 'Home Position')
PARS_16 = add_FloatParam(PARS_15, 'Setpoint', 'obs', 15, 0, 100, '#', 'actual setpoint')
PARS_17 = add_EnumParam(PARS_16, 'microsteps', 'rw', 16, 0, {s:f'{2**s} steps' for s in range(9)}, 'microstepping selection')
PARS_18 = add_IntParam(PARS_17, 'numbits', 'rw', 17, 1, 32, 'bit', 'number of bits per ssi transfer')

TCOL_ = add_EnumParam(PARS_10, 'enable', 'rw', 0, {2:'enabled', 1:'updating', 0:'ignored'}, {2:'enabled', 0:'ignored'},
                      'en/disable this table row')
TCOL = add_EnumParam(TCOL_, 'microsteps', 'rw', 9, 0, {s:f'{2**s} steps' for s in range(9)}, 'microstepping selection')

SFUNC_1 = add_SFunc(PARS_18, 'reset_to_factory_default', 128,
                    add_IntParam(0, 'unlock_value', 'rw', 0, -2**31, 2**31-1, '', 'Unlock value'),
                    0,
                    'resets everything to factory defaults'
                    )
SFUNC_2 = add_SFunc(SFUNC_1, 'home', 133, 0, 0, 'starts a homing cycle')
SFUNC_3 = add_SFunc(SFUNC_2, 'SetPosition', 137, PARS_15, 0, 'sets a new current position')
SFUNC_4 = add_SFunc(SFUNC_3, 'ContMove', 142, PARS_13, 0, 'Starts a continuous movement')
SFUNC_5 = add_SFunc(SFUNC_4, 'SetBits', 240, PARS_18, PARS_18, 'set and read-back current number of bits')
SFUNC_6 = add_SFunc(SFUNC_5, 'GetBits', 241, 0, PARS_18, 'read-back current number of bits')

TABLE_2 = add_Table(SFUNC_6, 'table2', addrof('table32'), 2, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, direct mapped')
TABLE_1 = add_Table(TABLE_2, 'table1', addrof('table32_act_row'), 1, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, with line select')
TABLE32 = add_Table(TABLE_1, 'table0', 1000, 0, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, via paramctlif')

ENUM_1 = add_Enums(0, {0: 'On', 1: 'Off'})
ENUM_2 = add_Enum(ENUM_1, 2, 'Moving')

VT_UINT32 = add_IntParam(0, 'ignored pname', 'rw', -1, 0, 2**32-1, '', 'unsigned 32 bit integer')

add_Device('di32', 0x1a04, addrof('dx32_target'), value_param=VT_UINT32, description='discrete input, 32 bit',
           aux=AUX24)
add_Device('pi32', 0x4008, addrof('px32_target'), unit='%', description='parameter output, 32 bit',
           aux=AUX24, params=TABLE32, errid=ERRID)
add_Device('po32', 0x500a, addrof('px32_value'), unit='%', description='parameter output, 32 bit',
           aux=AUX24, errid=ERRID, params=TABLE32)


PLC_DESCRIPTOR.description = add_String('simulation for testing zapf')
PLC_DESCRIPTOR.version = add_String('https://forge.frm2.tum.de/review/mlz/pils/zapf:v2.1-alpha')
PLC_DESCRIPTOR.author = add_String('anonymous\nauthor')


class Global(Globals):
    fMagic = Var(real, PILS_MAGIC, at='%MB0')
    iOffset = Var(word, INDEXER_OFFSET, at='%MB4')

    stIndexer = Var(ST_Indexer, at=f'%MB{INDEXER_OFFSET}')

    data = Var(DevicesLayout, at='%MB200')

    iCycle = Var(word, 0)


g = Global()

@program(
    nDevices = Var(word),
    devnum = Var(word),
    infotype = Var(word),
    itemp = Var(byte),
    tempofs = Var(word),
)
def Indexer(v):
    if g.fMagic != PILS_MAGIC:
        g.fMagic = PILS_MAGIC
    if g.iOffset != INDEXER_OFFSET:
        g.iOffset = INDEXER_OFFSET

    g.iCycle += 1

    if g.stIndexer.Request[[15]]:
        return

    req_num = g.stIndexer.Request & 0x7FFF

    data = g.stIndexer.Data
    memset(adr(data), 0, sizeof(data))

    if req_num == 32767:
        DEBUG_DESCRIPTOR.cycle = g.iCycle
        memcpy(adr(data), adr(DEBUG_DESCRIPTOR), sizeof(DEBUG_DESCRIPTOR))
    else:
        for i in range(DESCRIPTOR_SLOTS):
            if req_num + i < len(DESCRIPTOR_ARRAY):
                memcpy(adr(data) + i*DESCRIPTOR_SLOT_SIZE,
                       adr(DESCRIPTOR_ARRAY[req_num + i]),
                       DESCRIPTOR_SLOT_SIZE)

    g.stIndexer.Request[[15]] = 1


# helper
def advance32(value, target, status, usermin=0.0, usermax=100.0,
              warnmin=10.0, warnmax=90.0, speed=10.0):
    state = status >> 12
    reason = 0
    if state == 0:
        state = 1
    elif state in [1, 3, 7]:
        state = 1 if warnmin <= value <= warnmax else 3
    elif state == 5:
        state, reason = (6, 0) if usermin <= target <= usermax else (8, 1)
    elif state == 6:
        if value == target:
            state, reason = 1, 0
        value += max(min((target - value), speed / 20), -speed / 20)
    target = max(min(usermax, target), usermin)
    if value < warnmin:
        reason |= 4
    if value > warnmax:
        reason |= 8

    return value, target, (state << 12) | (reason << 8) | (1 << ((g.iCycle >> 9) & 7))


# same as above but for 32 bit status fields
def advance64(value, target, status, *args, **kwds):
    v, t, s = advance32(value, target, max(status >> 16, status & 0xFFFF), *args, **kwds)
    return v, t, s << 16 | s


# pylint: disable=too-many-return-statements
def handle_pctl(pctl, pvalue, valuestore, tablestore):
    pnum = pctl & 8191
    cmd = pctl >> 13
    if pctl & 0x8000:
        # nothing to do
        return pctl, pvalue
    elif not cmd:
        return 0x2000, pvalue
    if pnum in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
                17, 128, 133, 137, 142, 240, 241] or 1000 <= pnum <= 1099:
        # param exists, can at least be read
        if cmd == 1:  # DO_READ
            if 0 <= pnum <= 17:
                # return DONE and value
                return 0x8000 | pnum, valuestore[pnum]
            elif pnum == 240:
                return 0x8000 | pnum, valuestore[17]
            elif pnum == 241:
                return 0x8000 | pnum, valuestore[17]
            elif 1000 <= pnum <= 1099:
                # table access
                # return DONE and value
                return 0x8000 | pnum, float(tablestore[pnum - 1000])
            # invent value and DONE
            return 0x8000 | pnum, 0
        if cmd == 2:  # DO_WRITE
            if 0 <= pnum <= 17:
                valuestore[pnum] = pvalue
                return 0x2000 | pnum, pvalue  # -> read
            elif pnum == 240:
                valuestore[17] = pvalue
                return 0x8000 | pnum, valuestore[17]
            elif pnum == 241:
                return 0x8000 | pnum, valuestore[17]
            elif 128 <= pnum <= 239:
                # go to BUSY for some and ERR_RETRY for others
                if pnum in (128, 133):
                    return 0x6000 | pnum, pvalue
                return 0xE000 | pnum, pvalue  # ERR_RETRY
            elif 1000 <= pnum <= 1099:
                tablestore[pnum - 1000] = float(pvalue)
                return 0x2000 | pnum, pvalue  # -> read
            # no function, no storage -> ERR_READ_ONLY
            return 0xC000 | pnum, pvalue
        if cmd == 3:  # BUSY
            return 0x2000 | pnum, pvalue
        raise RuntimeError(f'handle_pctl({pctl}, {pvalue}, {valuestore}): '
                           'This should never happen!')
    else:
        # non exisiting param -> ERR_NO_IDX
        return 0xA000 | pnum, pvalue


def handle_dx16(value, target, status):
    if status >> 12 == 5:
        return target, 0x6000
    return value, 0x1000


def handle_dx32(value, target, status):
    v, s = handle_dx16(value, target, max(status >> 16, status & 0xFFFF))
    return v, s << 16 | s


@program(msg=Var(array(byte, 0, 80)), msglen=Var(byte, 0))
def Implementation(v):
    d = g.data

    d.dx32_target &= 0xFFFFFF
    d.dx32_value, d.dx32_estatus = handle_dx32(
        d.dx32_value, d.dx32_target, d.dx32_estatus
    )

    d.px32_value, d.px32_target, d.px32_estatus = advance64(
        d.px32_value, d.px32_target, d.px32_estatus,
        usermin = d.fax32_params[0], usermax = d.fax32_params[1],
        warnmin = d.fax32_params[2], warnmax = d.fax32_params[3],
        speed = d.fax32_params[12],
    )
    d.px32_nerrid = d.px32_estatus >> 15 if d.px32_estatus[[31]] else 0
    d.px32_pctl, d.px32_pvalue = handle_pctl(
        d.px32_pctl, d.px32_pvalue, d.fax32_params, d.table32
    )

    # handle table lines
    if d.table32_req_row != d.table32_act_row:
        # write back
        memcpy(
            adr(d.table32) + 40 * d.table32_act_row,
            adr(d.table32_line),
            sizeof(d.table32_line),
        )
        # avoid illegal values for req_row
        if not 0 <= d.table32_req_row <= 9:
            d.table32_req_row = d.table32_act_row
        # read from table
        memcpy(
            adr(d.table32_line),
            adr(d.table32) + 40 * d.table32_req_row,
            sizeof(d.table32_line),
        )
        d.table32_act_row = d.table32_req_row


@program()
def Main(v):
    # slow down simulation; this is both good for CPU load and for
    # simulating a slow connection for Zapf
    time.sleep(0.01)
    Indexer()
    Implementation()
